#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "UDCFSupport.h"
#include "UDCF.h"

#ifdef WITH_DMALLOC 
#include<dmalloc.h>
#endif                                                                          

/* ---- Global pointer for complete data contained in a UDCF file ---- */
UDCF* UDCFGlobal; 

/* ---- Prototyp for yyerror ---- */
extern int yyerror(char* msg);

char* saveStrDup(char* s)
{
 char *tmp;

 /* ---- Check for empty string ---- */
 if (s == (char *) NULL) return (char *) NULL;

 /* ---- Duplicate string ---- */
 tmp = strdup(s);

 /* ---- Test allocation ---- */
 if (tmp == (char *) NULL) {

  /* ---- Give an error ---- */
  yyerror("Can't allocate memory for strdup");

  /* ---- And terminate ---- */
  exit(EXIT_FAILURE);

 }

 /* ---- Return the pointer ---- */
 return tmp;

}

void increaseMix()
{
 static const long inc = 8;
 UDCFDive* dive = GetLastDive(); 

 /* ---- Check dive ---- */
 if (dive == (UDCFDive *) NULL) {

  /* ---- Give an error message ---- */
  yyerror("Invalid dive! Have to quit\n");
  
  /* ---- And terminate ---- */
  exit(EXIT_FAILURE);

 } 

 /* ---- Resize the array ---- */
 dive -> mixList = realloc(dive -> mixList, 
                           (dive -> mixSize + inc) * 
                           sizeof (UDCFMix)); 
 dive -> mixSize += inc;

 /* ---- Check realloc ---- */
 if (dive -> mixList == (UDCFMix *) NULL) {

  /* ---- Give an error message ---- */
  yyerror("Can't allocate memory for mixes array");

  /* ---- And terminate ---- */
  exit(EXIT_FAILURE);

 }

}

void increaseGroup()
{
 static const long inc = 4;
 
 /* ---- Resize the array ---- */
 UDCFGlobal -> groupList = realloc(UDCFGlobal -> groupList, 
                                   (UDCFGlobal -> groupSize + inc) * 
                                   sizeof (UDCFGroup)); 
 UDCFGlobal -> groupSize += inc;

 /* ---- Check realloc ---- */
 if (UDCFGlobal -> groupList == (UDCFGroup *) NULL) {

  /* ---- Give an error message ---- */
  yyerror("Can't allocate memory for groups array");

  /* ---- And terminate ---- */
  exit(EXIT_FAILURE);

 }

}

void increaseDive()
{
 static const long inc = 8;
 UDCFGroup* actGroup = GetLastRepGroup();

 /* ---- Check for valid group ---- */
 if (actGroup == (UDCFGroup *) NULL) {

  /* ---- Give an error an terminate ---- */
  yyerror("Can't increase dive-array! Illegal Group\n");
  exit(EXIT_FAILURE);

 }

 /* ---- Resize the array ---- */
 actGroup -> diveList = realloc(actGroup -> diveList,
                                (actGroup -> diveSize + inc) *
                                sizeof(UDCFDive));
 actGroup -> diveSize += inc;

 /* ---- Check the realloc ---- */
 if (actGroup -> diveList == (UDCFDive *) NULL) {

  /* ---- Give an error ---- */
  yyerror("Can't allocate memory for dives array");

  /* ---- And terminate ---- */
  exit(EXIT_FAILURE);
 
 }

}

void increaseSample()
{
 static const long inc = 64;
 UDCFDive* actDive = GetLastDive();

 /* ---- Resize the array ---- */
 actDive -> sampleList = realloc(actDive -> sampleList,
                                (actDive -> sampleSize + inc) *
                                sizeof(UDCFSample));
 actDive -> sampleSize += inc;

 /* ---- Check the realloc ---- */
 if (actDive -> sampleList == (UDCFSample *) NULL) {

  /* ---- Give an error ---- */
  yyerror("Can't allocate memory for samples array");

  /* ---- And terminate ---- */
  exit(EXIT_FAILURE);
 
 }
 
}

void CreateNewMix()
{
 UDCFDive *dive = GetLastDive();

 /* ---- Check for valid dive ---- */
 if (dive == (UDCFDive *) NULL) {

  /* ---- Give an error ---- */
  yyerror("Can't create an new mix because of missing dive\n");

  /* ---- Terminate ---- */
  exit(EXIT_FAILURE);

 }

 /* ---- Check if we have reached the end of array ---- */
 if ((dive -> mixIndex + 1) >= dive -> mixSize) {

  /* ---- Increase the array of mixes ---- */
  increaseMix();

 }

 /* --- Increase the index ---- */ 
 dive -> mixIndex++;
 
 /* --- Init the entry ---- */
 InitMix(GetLastMix());


}

void CreateNewRepGroup()
{
 /* ---- Check if we have reached the end of array ---- */
 if ((UDCFGlobal -> groupIndex + 1) >= UDCFGlobal -> groupSize) {

  /* ---- Increase the array ---- */
  increaseGroup();

 }

 /* ---- Increase the index ---- */
 UDCFGlobal -> groupIndex++; 

 /* ---- Init the new entry ---- */
 InitRepGroup(GetLastRepGroup());

}

void CreateNewDive()
{
 UDCFGroup* actGroup = GetLastRepGroup();

 /* ---- Check if size of array is suitable ---- */ 
 if ((actGroup -> diveIndex + 1) >= actGroup -> diveSize) {

  /* ---- Increase the array ---- */
  increaseDive();

 }

 /* ---- Increase the index ---- */
 actGroup -> diveIndex++;

 /* ---- Init the new entry ---- */
 InitDive(GetLastDive());

}

void CreateNewSample()
{
 UDCFDive* actDive = GetLastDive();

 /* ---- Check if size of array is suitable ---- */ 
 if ((actDive -> sampleIndex + 1) >= actDive -> sampleSize) {

  /* ---- Increase the array ---- */
  increaseSample();

 }

 /* ---- Increase the index (Last sample in actDive) ---- */
 actDive -> sampleIndex++;

 /* ---- Init the new sample ---- */
 InitSample(&(actDive -> sampleList[actDive -> sampleIndex]));

}

void InitRepGroup(UDCFGroup* group)
{
 /* ---- Init the list of dives ---- */
 group -> diveList = (UDCFDive* ) NULL; 
 group -> diveSize = 0;
 group -> diveIndex = -1; 

}

UDCF* GetProfile()
{
 /* ---- Simply return the global pointer ---- */
 return UDCFGlobal;

}

UDCFMix* GetLastMix()
{
 UDCFDive *dive = GetLastDive();

 /* ---- Check for valid dive ---- */
 if (dive == (UDCFDive *) NULL) {

  /* ---- Give an error and terminate ---- */
  yyerror("Can't get mix because of missing dive\n");

  /* ---- Terminate ---- */
  exit(EXIT_FAILURE);

 }

 /* ---- Check for valid mixes ---- */
 if (dive -> mixIndex <= -1) return (UDCFMix *) NULL;

 /* ---- Get the highest index and return a pointer to ---- */
 /* ---- the corresponding mix                         ---- */
 return &dive -> mixList[dive -> mixIndex];

}

bool IsCorrectMix(UDCFMix* mix)
{
#ifndef NDEBUG
 fprintf(stderr, "Check mix %s\n", mix -> mixName);
 fprintf(stderr, 
         "N2: %f O2: %f He: %f\n",
         mix -> n2Ratio, 
         mix -> o2Ratio, 
         mix -> heRatio);
#endif

 /* ---- Check whether the partial pressures sums up to 1.0 ---- */
 return (fabs(mix -> n2Ratio + mix -> o2Ratio + mix -> heRatio - 1.0) <= epsilon);

}

long FindMix(UDCFDive* dive, const char* mixName)
{
 long i;

 /* ---- Check for dive ---- */
 if (dive == (UDCFDive *) NULL) return C_DIVE_INVALID;

 /* ---- Check all mixes (Note: index of last mix is mixIndex!) ---- */
 for (i=0; i<= dive -> mixIndex; i++) {

  /* ---- Check if the this is the mix we are looking for ---- */
  if (strcmp(dive -> mixList[i].mixName, mixName) == 0) {

   /* ---- Found! Return the index ---- */
   return i;

  }

 }

 /* ---- No mix with name 'mixName' found ---- */
 return C_DIVE_INVALID;

}

UDCFGroup* GetLastRepGroup()
{

 /* ---- Check for valid entries ---- */
 if (UDCFGlobal -> groupIndex <= -1) return (UDCFGroup *) NULL;

 /* ---- Get the highest index and return a pointer to ---- */
 /* ---- corresponding group of dives                  ---- */
 return &UDCFGlobal -> groupList[UDCFGlobal -> groupIndex];

}

void InitMix(UDCFMix* mix)
{
 /* ---- Init the mix - name ---- */
 mix -> mixName = (char *) NULL;

 /* ---- Init pressure info ---- */
 mix -> pressureStart = C_DIVE_INVALID;
 mix -> pressureEnd = C_DIVE_INVALID;

 /* ---- Init the tank volume ---- */
 mix -> tankVolume = C_DIVE_INVALID;

 /* ---- Init the gas ratios ---- */
 mix -> heRatio = 0.0;
 mix -> o2Ratio = 0.0;
 mix -> n2Ratio = 0.0;

 /* ---- Init the acv (volume of gas needed per time) ---- */
 mix -> acv = 0.0;

}

void InitDive(UDCFDive* dive)
{

 /* ---- Init date and time info ---- */
 dive -> day    = C_DIVE_INVALID; 
 dive -> month  = C_DIVE_INVALID; 
 dive -> year   = C_DIVE_INVALID;
 dive -> minute = C_DIVE_INVALID; 
 dive -> hour   = C_DIVE_INVALID;

 /* ---- Init surface intervall ---- */
 dive -> surfaceInterval = C_DIVE_INVALID;

 /* ---- Init the temperature ---- */
 dive -> temperature = C_DIVE_INVALID;

 /* ---- Init diving place string ---- */
 dive -> place = (char *) NULL;

 /* ---- Init the altitude level ---- */
 dive -> altitude = C_DIVE_INVALID;

 /* ---- Init the meta data field ---- */
 dive -> prgData = (char *) NULL;

 /* ---- Init the specific gravitiy ---- */
 dive -> density = C_DIVE_INVALID;

 /* ---- Init user data field ---- */
 dive -> userDataType = C_DIVE_INVALID;
 dive -> userData = NULL;
 
 /* ---- Init profile data ---- */
 dive -> sampleList = (UDCFSample *) NULL;
 dive -> sampleSize = 0;
 dive -> sampleIndex = -1;

 /* ---- Init sample mode ---- */
 dive -> mode = M_DELTA;

 /* ---- Init the dynamic array of mixes used ---- */
 dive -> mixSize = 0;
 dive -> mixIndex = -1;
 dive -> mixList = (UDCFMix *) NULL;

}

UDCFDive* GetLastDive()
{
 UDCFGroup* highGroup = GetLastRepGroup();

 /* ---- Check for valid RepGroup entry ---- */
 if ((highGroup == (UDCFGroup *) NULL) || (highGroup -> diveIndex <= -1)) {

  /* ---- There exists no entry ---- */
  return (UDCFDive *) NULL;

 }

 /* ---- Return the last dive of RepGroup 'highGroup' ---- */
 return &(highGroup -> diveList[highGroup -> diveIndex]);

}

void InitSample(UDCFSample* sample) {

 /* ---- Init the sample ---- */
 sample -> type = ST_UNKNOWN;
 sample -> data.mark = 0;

}

#ifndef NDEBUG
/* ---- Print a dive (Debug-Mode) ---- */
void PrintDive(UDCFDive* dive)
{
 long i;
 
 fprintf(stderr, "Day: %d, Month: %d, Year %d\n", 
                 dive -> day, 
                 dive -> month, 
                 dive -> year);
 fprintf(stderr, "Minute: %d, Hour: %d\n", dive -> minute, dive -> hour);
  
 if (dive -> mode == M_DELTA) 
  fprintf(stderr, "Mode: Delta\n");
 else if (dive -> mode == M_TD) 
  fprintf(stderr, "Mode: Time/Depth\n");
 else
  fprintf(stderr, "Unknown Mode\n");

 fprintf(stderr, "SampleIndex %ld, SampleSize %ld\n", 
                 dive -> sampleIndex, 
                 dive -> sampleSize);

 for (i = 0; i <= dive -> sampleIndex; i++) {

  fprintf(stderr, "Sample %ld ", i);

  if (dive -> sampleList[i].type == ST_DEPTH) {

   fprintf(stderr, "Depth %f\n", dive -> sampleList[i].data.depth);
  
  } else if (dive -> sampleList[i].type == ST_TD) {

   fprintf(stderr, "XY (%f|%f)\n",  
                   dive -> sampleList[i].data.point.x, 
                   dive -> sampleList[i].data.point.y); 

  } else if (dive -> sampleList[i].type == ST_SAMPLETIME) {

   fprintf(stderr, "New SampleTime : %f\n",  
                   dive -> sampleList[i].data.sampleTime);
 
  } else if (dive -> sampleList[i].type == ST_MARK) {

   fprintf(stderr, "Mark ");
   switch (dive -> sampleList[i].data.mark) {
  
    case C_ASC:   fprintf(stderr,  "Ascent\n"); break;
    case C_DECO:   fprintf(stderr, "Deco\n"); break;
    case C_SURF:   fprintf(stderr, "Surf\n"); break;
    case C_BREATH: fprintf(stderr, "BREATH\n"); break;
 
    default: fprintf(stderr, "Unknown\n");

   }

  } else if (dive -> sampleList[i].type == ST_MIX) {

   fprintf(stderr, "Mix %s\n", dive -> mixList[dive -> sampleList[i].data.mix].mixName);

  } else if (dive -> sampleList[i].type == ST_TEMPERATURE) {

   fprintf(stderr, 
           "New temperature %6.2f\n", 
           dive -> sampleList[i].data.temperature);

  } else {

   fprintf(stderr, "Unknown sample type\n");
 
  }

 }

 /* ---- Flush stderr ---- */
 fflush(stderr);

}
#endif

int yywrap()
{
 /* ---- Send an EOF when reaching the end of file (no wrap-around!) ---- */
 return 1;
}
