%{

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#ifndef WIN32
#include <unistd.h>
#endif

#include "config.h"

#ifdef WITH_DMALLOC 
#include<dmalloc.h>
#endif

/* ---- List of tokens generated by bison ---- */
#include "parser.h"

/* ---- We don't need yyunput ---- */
#define YY_NO_UNPUT

/* ---- Hold the number of the first erroneous line ---- */
long errorLine; 

/* ----- Counter for line number ---- */
unsigned long lineno = 1;

/* ---- Protos ---- */
int yyerror(char *);

/* ---- Add char to metaData ---- */
void ConcatChar(char** ptr, long* size, long* index, char ch)
{
 const long incr = 64;

 /* ---- Check the size of the array ---- */
 if (((*index) + 1) >= *size) {

  /* ---- Resize the array ---- */
  (*size) += incr;
  *ptr = realloc(*ptr, (*size) * sizeof(char));

  /* ---- Check for success ---- */
  if ((*ptr) == (char*) NULL) {

   /* ---- Give an error and terminate ---- */
   yyerror("No enough memory for ConcatChar\n");
   exit(EXIT_FAILURE);  

  }

 }

 /* ---- Add the character ---- */
 (*index)++;
 (*ptr)[*index] = ch;

}

/* ---- Pointer, sizeof and index for external program data ---- */
char* prgData = (char *) NULL;
long prgSize = 0;
long prgIndex = -1;

/* ---- Pointer, sizeof and index for diving place string ---- */
char* placeData = (char *) NULL;
long placeSize = 0;
long placeIndex = -1;

%}

/* ---- Exclusive start states ---- */
%x PRGSTATE
%x PLACESTATE
%x STRSTATE
%x COMMENTSTATE
%x REALSTATE
%x SINTSTATE

%%

%{
#ifdef yywrap
#undef yywrap
#endif
%}

\n {lineno++;} 

[ \t] ; 

PROFILE       {return PROFILE;}
\<\/PROFILE\> {return END_PROFILE;}

\" {return DQUOTE;}

\<UNITS\>   {return UNITS;}
\<\/UNITS\> {return END_UNITS;}

Metric   {return METRIC;}
Imperial {return IMPERIAL;}

\<DEVICE\>                 {return DEVICE;}
\<\/DEVICE\>               {return END_DEVICE;}
\<VENDOR\> BEGIN STRSTATE; {return VENDOR;}
\<\/VENDOR\>               {return END_VENDOR;}
\<MODEL\> BEGIN STRSTATE;  {return MODEL;}
\<\/MODEL\>                {return END_MODEL;}
\<VERSION\>                {return UDCFVERSION;}
\<\/VERSION\>              {return END_UDCFVERSION;}

\<ADDINFO\>                  {return ADDINFO;}
\<\/ADDINFO\>                {return END_ADDINFO;}
\<PERSINFO\> BEGIN STRSTATE; {return PERSINFO;}
\<\/PERSINFO\>               {return END_PERSINFO;}
\<TOTALDIVES\>               {return TOTALDIVES;}
\<\/TOTALDIVES\>             {return END_TOTALDIVES;}
\<SERIALID\> BEGIN STRSTATE; {return SERIALID;}
\<\/SERIALID\>               {return END_SERIALID;}

\<GASES\>                       {return GASES;}
\<\/GASES\>                     {return END_GASES;}
\<MIX\>                         {return MIX;}
\<\/MIX\>                       {return END_MIX;}
\<MIXNAME\> BEGIN STRSTATE;     {return MIXNAME;}
\<\/MIXNAME\>                   {return END_MIXNAME;}
\<TANK\>                        {return TANK;}
\<\/TANK\>                      {return END_TANK;}
\<PSTART\> BEGIN REALSTATE;     {return PSTART;}
\<\/PSTART\>                    {return END_PSTART;}
\<PEND\> BEGIN REALSTATE;       {return PEND;}
\<\/PEND\>                      {return END_PEND;}
\<TANKVOLUME\> BEGIN REALSTATE; {return TANKVOLUME;}
\<\/TANKVOLUME\>                {return END_TANKVOLUME;}
\<N2\> BEGIN REALSTATE;         {return N2;}
\<\/N2\>                        {return END_N2;}
\<O2\> BEGIN REALSTATE;         {return O2;}
\<\/O2\>                        {return END_O2;}
\<HE\> BEGIN REALSTATE;         {return HE;}
\<\/HE\>                        {return END_HE;}
\<ACV\> BEGIN REALSTATE;        {return ACV;}
\<\/ACV\>                       {return END_ACV;}

\<REPGROUP\>   {return REPGROUP;}
\<\/REPGROUP\> {return END_REPGROUP;}

\<DIVE\>                                  {return DIVE;}
\<\/DIVE\>                                {return END_DIVE;}
\<DELTA\> BEGIN REALSTATE;                {return DELTA;}
\<\/DELTA\>                               {return END_DELTA;}
\<DATE\>                                  {return DATE;}
\<\/DATE\>                                {return END_DATE;}
\<DAY\>                                   {return DAY;}
\<\/DAY\>                                 {return END_DAY;}
\<MONTH\>                                 {return MONTH;}
\<\/MONTH\>                               {return END_MONTH;}
\<YEAR\>                                  {return YEAR;}
\<\/YEAR\>                                {return END_YEAR;}
\<TIME\>                                  {return TIME;}
\<\/TIME\>                                {return END_TIME;}
\<HOUR\>                                  {return HOUR;}
\<\/HOUR\>                                {return END_HOUR;}
\<MINUTE\>                                {return MINUTE;}
\<\/MINUTE\>                              {return END_MINUTE;}
\<TEMPERATURE\> BEGIN REALSTATE;          {return TEMPERATURE;}
\<\/TEMPERATURE\>                         {return END_TEMPERATURE;}
\<DENSITY\> BEGIN REALSTATE;              {return DENSITY;}
\<\/DENSITY\>                             {return END_DENSITY;}
\<ALTITUDE\> BEGIN REALSTATE;             {return ALTITUDE;}
\<\/ALTITUDE\>                            {return END_ALTITUDE;}
\<TIMEDEPTHMODE\/\>                       {return TIMEDEPTHMODE;}
\<TIMEDEPTHMODE\>[ \t]*<\/TIMEDEPTHMODE\> {return TIMEDEPTHMODE;}
\<DELTAMODE\/\>                           {return DELTAMODE;}
\<DELTAMODE\>[ \t]*<\/DELTAMODE\>         {return DELTAMODE;}
\<SAMPLES\>                               {return SAMPLES;}
\<\/SAMPLES\>                             {return END_SAMPLES;}
\<D\> BEGIN REALSTATE;                    {return DEPTH;}
\<\/D\>                                   {return END_DEPTH;}
\<T\> BEGIN REALSTATE;                    {return T;}
\<\/T\>                                   {return END_T;}
\<SWITCH\> BEGIN STRSTATE;                {return SWITCH;}
\<\/SWITCH\>                              {return END_SWITCH;}
\<SURFACEINTERVAL\> BEGIN SINTSTATE;      {return SURFACEINTERVAL;}
\<\/SURFACEINTERVAL\>                     {return END_SURFACEINTERVAL;}

\<PLACE\> BEGIN PLACESTATE; {
                             /* ---- Init dynamic prgData array ---- */
                             placeData = (char *) NULL;
                             placeSize = 0;
                             placeIndex = -1;
                                         
                             /* ---- Return the token ---- */
                             return PLACE;
                
                            }

<PLACESTATE>\n {ConcatChar(&placeData, &placeSize, &placeIndex, yytext[0]); lineno++;} 
<PLACESTATE>.  {ConcatChar(&placeData, &placeSize, &placeIndex, yytext[0]);}
<PLACESTATE>\<\/PLACE\> BEGIN INITIAL; { 
                                        /* ---- Terminate the string ---- */
                                        ConcatChar(&placeData, 
                                                   &placeSize, 
                                                   &placeIndex, 
                                                   '\0'); 
                           
                                          /* ---- Return the complete prgData ---- */
                                          yylval.string = placeData; 
                                          return END_PLACE;
                                         }


\<PROGRAM\> BEGIN PRGSTATE; { 
                             /* ---- Init dynamic prgData array ---- */
                             prgData = (char *) NULL;
                             prgSize = 0;
                             prgIndex = -1;
                                         
                             /* ---- Return the token ---- */
                             return PRG;
                            }
<PRGSTATE>\n {ConcatChar(&prgData, &prgSize, &prgIndex, yytext[0]); lineno++;} 
<PRGSTATE>.  {ConcatChar(&prgData, &prgSize, &prgIndex, yytext[0]);}
<PRGSTATE>\<\/PROGRAM\> BEGIN INITIAL; { 
                                        /* ---- Terminate the string ---- */
                                        ConcatChar(&prgData, 
                                                   &prgSize, 
                                                   &prgIndex, 
                                                   '\0'); 
                           
                                        /* ---- Return the complete prgData ---- */
                                        yylval.string = prgData; 
                                        return END_PRG;
                                       }


\<!-- BEGIN COMMENTSTATE;
<COMMENTSTATE>\n ; 
<COMMENTSTATE>. ;
<COMMENTSTATE>--\> BEGIN INITIAL; 

UDCF {return MARKER;}

\< {return BRACK_OPEN;}
\> {return BRACK_CLOSE;}
=  {return EQ;}

\<ALARM\>ASCENT\<\/ALARM\>  {return ASCENT;}
\<ALARM\>DECO\<\/ALARM\>    {return DECO;}
\<ALARM\>SURFACE\<\/ALARM\> {return SURFACE;}
\<ALARM\>BREATH\<\/ALARM\>  {return BREATH;}
\<ALARM\>RBT\<\/ALARM\>     {return RBT;}  
\<ALARM\>ERROR\<\/ALARM\>   {return ERROR;}  

\<WORK\>   {return WORK;}  
\<\/WORK\> {return END_WORK;}  

-?[0-9]+ {

 long erg;

 /* ---- Convert string in number ----- */
 errno = 0;
 erg = strtol(yytext, (char **) NULL, 10);

 /* ---- Check for success ---- */
 if (errno != 0) {

  /* ---- There were an error ---- */
  yyerror("Can't convert to integer");

 }

 /* ---- Return value of token ---- */
 yylval.ival = erg;

 /* ---- Return the right token ---- */
 return INTEGER;

}

<SINTSTATE>(((([0-9]*\.[0-9]+)([eE][-+]?[0-9]+)?)|[0-9]+)|(infinity)) BEGIN INITIAL; {

 double erg;

 /* ---- Check for keyword infinity ---- */
 if (strcmp(yytext,"infinity") != 0) { 

  /* ---- Convert string in double ----- */
  erg = (double) strtod(yytext, (char **) NULL);

  /* ---- Everything under control? ---- */
  if (errno != 0) {

   /* ---- There were an error ---- */
   yyerror("Can't convert to double");

  }

  /* ---- Return value of token ---- */
  yylval.fval = erg;
  
  /* ---- Return the token ---- */
  return REAL;

 } else {

  /* ---- We found the keyword infinity ---- */
  return INFINITY;

 }

}

<REALSTATE>((([0-9]*\.[0-9]+)([eE][-+]?[0-9]+)?)|[0-9]+) BEGIN INITIAL; {

 double erg;

 /* ---- Signed real can be done by :                             ---- */
 /* ---- -?((([0-9]+)|([0-9]*\.[0-9]+)([eE][-+]?[0-9]+)?)|[0-9]+) ---- */

 /* ---- Convert string in double ----- */
 errno = 0;
 erg = (double) strtod(yytext, (char **) NULL);

 /* ---- Everything under control? ---- */
 if (errno != 0) {

  /* ---- There were an error ---- */
  yyerror("Can't convert to double");

 }

 /* ---- Return value of token ---- */
 yylval.fval = erg;

 /* ---- Return the token ---- */
 return REAL;

}

<STRSTATE>[a-zA-Z0-9\+\-\.@.,:;'=&\!$\~\(\)\*\ ]+ BEGIN INITIAL; {

 char* erg = (char*) NULL;
 
 /* ---- Allocate enough memory ---- */
 erg = malloc(strlen(yytext) + 1); 

 /* ---- Check for success ---- */
 if (erg == (char*) NULL) {

  /* ---- Call yyerror ---- */
  yyerror("Can't allocate memory for string to parse");

  /* ---- If yyerror does not terminate -> terminate ---- */
  exit(EXIT_FAILURE);

 }

 /* ---- Copy the string ---- */
 (void) strcpy(erg, yytext);
 yylval.string = erg;

 /* ---- Return the token ---- */
 return STRING;

}

%%

void yyfreebuffer()
{

#ifdef HAVE_FLEX

#ifndef NDEBUG
 fprintf(stderr, "Free memory of flex's buffers\n");
#endif

 /* ---- Free internal memory ---- */
 yy_delete_buffer(yy_current_buffer);
#endif

}

void resetLineNo()
{
 /* ---- Reset to first line number ---- */
 lineno = 1;
}

int yyerror(char* msg)
{
 /* ---- Tell line number of the erroneous line ---- */
 errorLine = lineno;

#ifndef NDEBUG
 fprintf(stderr, "Error (%s) in Line %ld\n", msg, lineno);
#endif

 /* ---- Free internal buffers ---- */
 yyfreebuffer();

 /* ---- Return 0 as an sign of failure ---- */
 return 0;

}
