%{
#include <stdlib.h> 

#include "UDCFSupport.h"

/* ---- Prototyp for yyerror and yylex ---- */
extern int yyerror(char* msg);
extern int yylex(void);

%}

/* ---- Datatyp containing the value of a token ---- */
%union {
        double fval;
        long ival;
        char* string;
       }

/* ---- Token for file footer ---- */
%token PROFILE
%token END_PROFILE

/* ---- Tokens for the Units-section ---- */
%token UNITS END_UNITS
%token METRIC IMPERIAL

/* ---- Tokens for the Device-sections ---- */
%token DEVICE END_DEVICE
%token VENDOR END_VENDOR
%token MODEL END_MODEL
%token UDCFVERSION END_UDCFVERSION

/* ---- Tokens for the additional-section ---- */
%token ADDINFO END_ADDINFO
%token PERSINFO END_PERSINFO
%token TOTALDIVES END_TOTALDIVES
%token SERIALID END_SERIALID

/* ---- Tokens for the gases-section ---- */
%token GASES END_GASES
%token MIX END_MIX
%token MIXNAME END_MIXNAME
%token TANK END_TANK
%token PSTART END_PSTART
%token PEND END_PEND
%token TANKVOLUME END_TANKVOLUME
%token N2 END_N2
%token O2 END_O2
%token HE END_HE
%token ACV END_ACV

/* ---- Rep-group section ---- */
%token REPGROUP END_REPGROUP

/* ---- Dive-token ---- */
%token DIVE END_DIVE
%token DELTA END_DELTA
%token DATE END_DATE
%token DAY END_DAY
%token MONTH END_MONTH
%token YEAR END_YEAR
%token TIME END_TIME
%token MINUTE END_MINUTE
%token HOUR END_HOUR
%token TEMPERATURE END_TEMPERATURE
%token DENSITY END_DENSITY
%token ALTITUDE END_ALTITUDE
%token TIMEDEPTHMODE
%token DELTAMODE
%token SAMPLES END_SAMPLES
%token DEPTH END_DEPTH
%token T END_T
%token SWITCH END_SWITCH
%token SURFACEINTERVAL END_SURFACEINTERVAL
%token <ival> INFINITY

%token PLACE 
%token <string> END_PLACE

%token PRG 
%token <string> END_PRG

/* ---- Token for fileformat marker ---- */
%token MARKER

/* ---- Tokens for brackets ---- */
%token BRACK_OPEN BRACK_CLOSE

/* ---- Token for = ---- */
%token EQ

/* ---- Token for " ---- */
%token DQUOTE

/* ---- Tokens for notes during dive ---- */
%token <fval> SURFACE ASCENT DECO BREATH RBT ERROR 
%token WORK END_WORK

/* ---- Tokens for numbers and strings (With type parameters) ---- */
%token <ival>   INTEGER
%token <fval>   REAL
%token <string> STRING

/* ---- Set type of nonterminals ---- */
%type <ival> day_desc 
%type <ival> month_desc 
%type <ival> year_desc 

%type <ival> hour_desc 
%type <ival> minute_desc 

%type <fval> press_in 
%type <fval> press_out 

%type <fval> secs 
%type <fval> intervaldesc 

%type <fval> delta 
%type <ival> version 

%type <fval> sample_list_delta 
%type <fval> sampledesc_delta  
%type <fval> sample_list_timedepth 
%type <fval> sampledesc_timedepth
%type <fval> point
%type <fval> depth 
%type <fval> markers

%type <string> mixname 
%type <fval> n2_ratio 
%type <fval> o2_ratio 
%type <fval> he_ratio 

/* ---- Start symbol ---- */
%start profile

%%

/* ---- Description of the profile format ---- */
profile: 
        profile_header
        units_section
        device_section
        additional_section
        repgroups
        END_PROFILE

/* ---- Detailed parsing of header ---- */
profile_header:
               BRACK_OPEN
               PROFILE
               MARKER
               EQ
               DQUOTE
               version
               DQUOTE 
               BRACK_CLOSE  
version: INTEGER {GetProfile() -> version = $1;}

/* ---- The section containing the units ---- */
units_section:
              UNITS
              unitsymbol 
              END_UNITS
unitsymbol:
	     METRIC   {GetProfile() -> measure = metric;}
           | IMPERIAL {GetProfile() -> measure = imperial;}          


/* ---- Grammar for hardware info ---- */
device_section:
               DEVICE
               VENDOR
               vendorString  
               END_VENDOR
               MODEL 
               modelString
               END_MODEL
               UDCFVERSION
               versionID
               END_UDCFVERSION
               END_DEVICE
vendorString: STRING {GetProfile() -> vendor = saveStrDup($1); free($1);}    
modelString: STRING {GetProfile() -> model = saveStrDup($1); free($1);}
versionID: INTEGER {GetProfile() -> driverVersion = $1;}

/* ---- Grammar for section containing additional information ---- */
additional_section:
                   ADDINFO
                   personal_desc
                   totalDives_desc
                   IDString_desc
                   END_ADDINFO
                  |
                  ;
personal_desc: 
               PERSINFO 
               STRING {GetProfile() -> personalInfo = saveStrDup($2); free($2);}
               END_PERSINFO
              |
              ;    
totalDives_desc: 
                 TOTALDIVES
                 INTEGER {GetProfile() -> totalDives = $2;}
                 END_TOTALDIVES
                |
                ;
IDString_desc: 
               SERIALID  
               STRING {GetProfile() -> serialID = saveStrDup($2); free($2);}
               END_SERIALID
              |
              ;

/* ---- Grammar for description of rep-groups ----- */
repgroups: 
             repgroups repgroupdesc
           | repgroupdesc

repgroupdesc:
              REPGROUP {CreateNewRepGroup();}
              divelist
              END_REPGROUP
divelist:
	    divelist dive_desc 
          | dive_desc 

/* ---- Description of a single dive ---- */
dive_desc:
	   DIVE {CreateNewDive();}
           place_desc
           date
           time
           surface_interval
           temperature
           density_desc
           alt_info
           gases_section
           prg_data
           samples
           END_DIVE

/* ---- Description of diving place ---- */
place_desc:
             PLACE
             END_PLACE {GetLastDive() -> place = saveStrDup($2); free($2);}
            |
	    ;

/* ---- Grammar for a date & time ---- */
date: 
      DATE 
      YEAR 
      year_desc
      END_YEAR 
      MONTH 
      month_desc
      END_MONTH 
      DAY
      day_desc 
      END_DAY 
      END_DATE
day_desc:   INTEGER {GetLastDive() -> day = $1;}
month_desc: INTEGER {GetLastDive() -> month = $1;}
year_desc:  INTEGER {GetLastDive() -> year = $1;}

time: 
     TIME
     HOUR 
     hour_desc 
     END_HOUR
     MINUTE 
     minute_desc 
     END_MINUTE
     END_TIME
hour_desc:   INTEGER {GetLastDive() -> hour = $1;}
minute_desc: INTEGER {GetLastDive() -> minute = $1;}


/* ---- Description of lowest temperature ---- */
temperature:
             TEMPERATURE temp END_TEMPERATURE
            |
            ;
temp: REAL {GetLastDive() -> temperature = $1;}

/* ---- Decription for density of water (Metric: in kg/m^3) ---- */
density_desc: DENSITY
              REAL {GetLastDive() -> density = $2;}
              END_DENSITY

/* ---- Description of the altitude ---- */
alt_info:
          ALTITUDE level END_ALTITUDE
level: REAL {GetLastDive() -> altitude = $1;}

/* ---- Grammar for a surfaceinterval in seconds ---- */
surface_interval: 
                  SURFACEINTERVAL 
                  intervaldesc 
                  END_SURFACEINTERVAL 
intervaldesc:   secs 
	      | INFINITY {GetLastDive() -> surfaceInterval = C_NAI;}
secs: REAL {GetLastDive() -> surfaceInterval = $1;}

/* ---- Section describing the mixes used (must not exists) ---- */
gases_section:
               GASES
               mix_list
               END_GASES

mix_list:   mix_desc mix_list
          | mix_desc
mix_desc: 
	  MIX {CreateNewMix();}
          MIXNAME mixname END_MIXNAME
          tank_info
          O2 o2_ratio END_O2
          N2 n2_ratio END_N2
          HE he_ratio END_HE
          acv_desc
          END_MIX {if(!IsCorrectMix(GetLastMix())) 
                    yyerror("Warning: Incorrect mix found!");}

mixname:  STRING {GetLastMix() -> mixName = saveStrDup($1); free($1);}
n2_ratio: REAL   {GetLastMix() -> n2Ratio = $1;}
o2_ratio: REAL   {GetLastMix() -> o2Ratio = $1;}
he_ratio: REAL   {GetLastMix() -> heRatio = $1;}

acv_desc: 
           ACV
           acv_value
           END_ACV
         |
         ;
acv_value: REAL {GetLastMix() -> acv = $1; }

/* ---- Description of in/out pressure ---- */
tank_info:
           TANK
           tank_volume_info 
           PSTART 
           press_in 
           END_PSTART
           press_end_desc
           END_TANK 
	  |
          ;

press_end_desc: 
                PEND 
                press_out 
                END_PEND
               |
               ;

press_in: REAL {GetLastMix() -> pressureStart = $1;}
press_out: REAL {GetLastMix() -> pressureEnd = $1;}

tank_volume_info:  TANKVOLUME
                   volume
                   END_TANKVOLUME
                  |
                  ;
volume: REAL {GetLastMix() -> tankVolume = $1;}

/* ---- Data section for application specific data ---- */
prg_data:
           PRG
           END_PRG {GetLastDive() -> prgData = saveStrDup($2); free($2);}
          |
	  ;

/* ---- Format of depth - samples ---- */
samples:
          delta_mode_marker
          SAMPLES
          delta_marker switch_desc sample_list_delta
          END_SAMPLES
        | timedep_mode_marker
          SAMPLES
          switch_desc sample_list_timedepth
          END_SAMPLES

/* ---- Delta mode ---- */
delta_mode_marker : DELTAMODE {GetLastDive() -> mode = M_DELTA;}

/* ---- Depth|Time mode ---- */
timedep_mode_marker: TIMEDEPTHMODE {GetLastDive() -> mode = M_TD;}

sample_list_delta:
                     sampledesc_delta sample_list_delta
		   | sampledesc_delta

sample_list_timedepth:
                        sampledesc_timedepth sample_list_timedepth
                      | sampledesc_timedepth

sampledesc_delta:   depth 
		  | sampletemperature {}
                  | switch_desc {}
                  | markers

sampledesc_timedepth:   point
		      | sampletemperature {}
	              | switch_desc {}
                      | markers

sampletemperature:
                   TEMPERATURE 
                   REAL      
		   END_TEMPERATURE {
                                    UDCFDive *lastDive = GetLastDive();
  
                                    /* ---- Insert a new temperatur - sample ---- */
                                    CreateNewSample();
                                    lastDive -> 
				     sampleList[lastDive -> 
                                      sampleIndex].type = ST_TEMPERATURE;
                                    lastDive -> 
                                     sampleList[lastDive -> 
				      sampleIndex].data.temperature = $2;

                 		   }

delta_marker:
              DELTA delta END_DELTA

delta: 
       REAL {
             UDCFDive *lastDive = GetLastDive();

             /* ---- Insert a delta-switch marker ---- */
             CreateNewSample();
             lastDive -> 
              sampleList[lastDive -> sampleIndex].type = ST_SAMPLETIME;
             lastDive -> 
              sampleList[lastDive -> sampleIndex].data.sampleTime = $1;

            }

markers:
         SURFACE {
                  UDCFDive *lastDive = GetLastDive();

                  /* ---- Insert surface - marker  ---- */
                  CreateNewSample();
                  lastDive -> 
                   sampleList[lastDive -> sampleIndex].type = ST_MARK;
                  lastDive -> 
                   sampleList[lastDive -> 
                    sampleIndex].data.mark = C_SURF;
                 }
        | ASCENT {
                  UDCFDive *lastDive = GetLastDive();

                  /* ---- Insert new slow - marker ---- */
                  CreateNewSample();
                  lastDive -> 
                   sampleList[lastDive -> sampleIndex].type = ST_MARK;
                  lastDive -> 
                   sampleList[lastDive -> 
                    sampleIndex].data.mark = C_ASC;
               }
         | DECO {
                 UDCFDive *lastDive = GetLastDive();

                 /* ---- Insert new deco - marker ---- */
                 CreateNewSample();
                 lastDive -> 
                  sampleList[lastDive -> sampleIndex].type = ST_MARK;
                 lastDive -> 
                  sampleList[lastDive -> 
                   sampleIndex].data.mark = C_DECO;
                }

         | BREATH {
                   UDCFDive *lastDive = GetLastDive();

                   /* ---- Insert new "out-of-breath" - marker ---- */
                   CreateNewSample();
                   lastDive -> 
                    sampleList[lastDive -> sampleIndex].type = ST_MARK;
                   lastDive -> 
                    sampleList[lastDive -> 
                     sampleIndex].data.mark = C_BREATH;
                  }

         | RBT {
                 UDCFDive *lastDive = GetLastDive();

                 /* ---- Insert new "remaining bottom time" - marker ---- */
                 CreateNewSample();
                 lastDive -> 
                  sampleList[lastDive -> sampleIndex].type = ST_MARK;
                 lastDive -> 
                  sampleList[lastDive -> 
                   sampleIndex].data.mark = C_RBT;
               }

         | ERROR {
                  UDCFDive *lastDive = GetLastDive();

                  /* ---- Insert new "general error" - marker ---- */
                 CreateNewSample();
                 lastDive -> 
                  sampleList[lastDive -> sampleIndex].type = ST_MARK;
                 lastDive -> 
                  sampleList[lastDive -> 
                   sampleIndex].data.mark = C_ERROR;
               }

         | workdesc {
                     UDCFDive *lastDive = GetLastDive();

                     /* ---- Insert new "Work"-sample ---- */
                     CreateNewSample();
                     lastDive -> sampleList[lastDive -> sampleIndex].type = ST_WORK;

                    }

depth:  
       DEPTH REAL END_DEPTH {
                             UDCFDive *lastDive = (UDCFDive *) NULL;

                             /* ---- Create space for new sample ---- */
                             CreateNewSample(); 

                             /* --- Get pointer to last sample ---- */
                             lastDive = GetLastDive();

                             /* ---- Insert the new sample ---- */
                             lastDive -> 
                              sampleList[lastDive -> 
                                          sampleIndex].type = ST_DEPTH;
                             lastDive -> 
                              sampleList[lastDive -> 
                                          sampleIndex].data.depth = $2;

                            }

point:  
       T REAL END_T
       DEPTH REAL END_DEPTH {
                             UDCFDive *lastDive = (UDCFDive *) NULL;
                      	    
                             /* ---- Create space for new sample ---- */
                             CreateNewSample(); 
 		     	    
                             /* --- Get pointer to last sample ---- */
                             lastDive = GetLastDive();
		     	    
                             /* ---- Insert the new sample ---- */
                             lastDive -> 
                              sampleList[lastDive -> 
                               sampleIndex].type = ST_TD;
                             lastDive -> 
                              sampleList[lastDive -> 
                               sampleIndex].data.point.x = $2;
                             lastDive -> 
                              sampleList[lastDive -> 
                               sampleIndex].data.point.y = $5;
		      
                             }

switch_desc:  
             SWITCH 
             changedMix 
             END_SWITCH 

changedMix: 
            STRING {
                    UDCFDive* lastDive = GetLastDive();
                    long mix = FindMix(lastDive, $1); 

                    /* ---- Free the name of the mix ---- */
                    free($1);
		    
                    /* ---- Check for a valid mix ---- */
                    if (mix == C_DIVE_INVALID) {
      		    
                     /* ---- Write a warning ---- */
                     yyerror("Unknown mix!");

    
		    }
		    
                    /* ---- Insert a gas switch marker ---- */
                    CreateNewSample();
                    lastDive -> 
                     sampleList[lastDive -> 
                      sampleIndex].type = ST_MIX;
		                         lastDive ->
                     sampleList[lastDive -> sampleIndex].data.mix =
                     mix; 

                    }

workdesc: WORK 
          REAL 
          END_WORK {
                    UDCFDive *lastDive = GetLastDive();

                    /* ---- Set the work-load ---- */
                    lastDive -> 
                     sampleList[lastDive -> sampleIndex].data.work = $2;

                   }

%%
